/\badvice\b/i
/\badvisory\b/i
/\baviso\b/i
/\bdocuments?\b/i
/\bcontract\b/i
/\bjustificante pago\b/i
/\bnotice\b/
/\borden de litigio\b/i
/\border\b/i
/\bpago\b/i
/\bpayments?\b/i
/\bRFQ\b/
/\bshipment\b/
/\bshipping\b/
/\bquotation\b/i
